/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*****************************************************************************
  program name: lm_pval.c

  Program to do posterior IBD probs
  The state (X) consists of segregation indicators

  X determines the IBD pattern among the specified descendant gametes.

  The data, Y, is the multilocus genotypes of observed individuals in the
  pedigree.

  lm_pval implemented, based on Feb 2003 Autozyg program lm_auto.c
                                                                EAT Apr 2003
  Modified to use globals identifying affected individuals,
  component info                                                MAM Jun 2003

  The realized probability distributions are scores according to one
     of several ibd measures and a p-value of a test for linkage is
     calculated at each marker locus, and a multiple-testing corrected
     one based on the maximum over loci.

  Modified to do correct fuzzy-pval dsns (per Geyer & Thompson) EAT Oct 2004
          First get null-null dsn quantiles accurate to .001 (0.1%)
          Then score marker loci, as generated, against this dsn.
  Generalized to use any ibd measure (currently max_fgl, spair, robdom)
    and to print out quantiles of the fuzzy-p distribution      EAT July 2005

  This version replaces the obsolete lm_pval of MORGAN 2.5 & 2.6
    for MORGAN 2.7

******************************************************************************/

/**
 * @defgroup lm_pval lm_pval
 * @ingroup autozyg
 * @{
 */

#define PNAM  "lm_pval"

#include "autozyg.h"

int main (int argc, char *argv[])
{
   int   j;
   int   nitnull, ntpos;
   int   t, c, h, jj;
   int   nsw;
   int ***recounts;                     /* [nc][nl][6] counts of
                                         *             recombinations     */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   double **prob;

   Measure scoraff;

   double **hdsn;                       /* [npos+1][nitnull]
                                         * list of null imputation scores
                                         * at all evaluatio positions     */
   double **fzpval;                     /* [npos+1][nnull]
                                         * summed over cmpt [c]           */

   double **qndsn;                      /* quantiles of the null dsn
                                         * sampled  [npos+1][nnull]
                                         * F^{-1}((k+1)/nnull)  is element k  */
   double *mean_stat;                   /* [npos+1] vector of trad stats  */
   int   nnull = 1000;                  /* number of quantiles for null  */
   int   id0;                           /* indicator for id null marginals */
   int   decr;                          /* number of realizations per nnull */
   int   tot_count = 0;                 /* score by component and locus   */

   FILE *oSeedF;

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   /* number of ibd score positions -- currently score only at markers */
   ntpos = NMarks;
   /* number of null realizations   -- currently set to the same as MCMC Its */
   nitnull = NIt;
   /* indicator of id null marginals -- will always be 1 in this prog    */
   id0 = 1;

   /* set up ibd measure: any of these 3 should currently work           */

   if (MaxFGLOn)
      scoraff = max_fgl;                /* default: used in current Gold2 */
   else if (RobdomOn)
      scoraff = robdom;
   else if (SpairsOn)
      scoraff = spair;

   setup_ibdmeas (scoraff);

   mean_stat = malloc ((ntpos + 1) * sizeof (double));
   for (j = 0; j <= ntpos; ++j)
      mean_stat[j] = 0.0;

   /* First generate the null-null distribution quantiles,      EAT  Oct 2004 */
   /*   This will be fixed, and any future sampling of null locus irrelevant  */
   /*    the 0 slot of the hdsn will be used to store the max over loci  */
   /*   For convenience, NIt iterations are used, but iid not mcmc            */

   /* Now realize segs and set hdsn from nitnull nulls */
   hdsn = alloc_dbl2d (ntpos + 1, nitnull);

   for (t = 0; t < nitnull; ++t)
   {
      /* realise the initial switches */

      realise (NMarks, xecomo, nsw);

      /* set founder genome labels (gg0, gg1) from switches */

      for (j = 1; j <= NMarks; ++j)
         set_FGL (j);

/* score the ibd measure -- store the value  */
      multi_gen_score (t, scoraff, ntpos, hdsn);
   }

/* convert dsn to 0.1 % quantiles */

   qndsn = alloc_dbl2d (ntpos + 1, nnull + 1);

   decr = nitnull / nnull;
   for (j = 0; j <= ntpos; ++j)
      make_quantiles (nnull, decr, hdsn[j], qndsn[j]);

/* free the full set of null dsn realizations: only the quantiles are needed */
   free_2d ((void *) &hdsn);

/* now print quantiles for check -- at some large interval                   */
   printf (" Now print null quantiles\n");
   fflush (stdout);
   pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
   fflush (stdout);

/* if marginals should be ident distributed (id0=1), then merge quantiles    */
   if ((id0) && (NMarks > 1))
   {
      compress_marginals (1, NMarks, nnull, qndsn);

/* now print quantiles again for check -- at some large interval             */
      printf (" Now print merged null quantiles\n");
      fflush (stdout);
      pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
      fflush (stdout);
   }

/* allocate and initialize log-seg and log-pen stuff */
/* probabilities now computed separately by component  */

   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

   for (c = 0; c < n_compnts; ++c)
   {
      for (h = 0; h <= NMarks; ++h)
      {
         tot_log_prob_seg[c][h] = 0.0;
         tot_log_prob_pen[c][h] = 0.0;
      }
      for (j = 0; j <= NMarks; ++j)
      {
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
      }
   }

/* allocate and zero the fuzzy-pvals   */
   fzpval = alloc_dbl2d (ntpos + 1, nnull + 1);

/* now intialise inheritamce  */
/*   prob = setupInherit (0);      revert to 0 for a valgrind check */
   prob = setupInherit (1);

   alloc_logprob ();

   prntStatusBMC ();
   /* now do burn-in iterations */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   fflush (stdout);
   for (t = 1; t <= NBrnIt; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burnin iteration %d\n", t);
         fflush (stdout);
      }
      sampling (t, 1, NMarks, nsw, nFGL, prob); /* arg2 =1: EAT 3/2005 */
   }

   /* now do MCMC sampling, and score results */

   printf (" Now starting %d main MCMC iterations \n", NIt);
   fflush (stdout);
   for (t = 0; t < NIt; ++t)
   {
      /* MCMC sampling step or scan */
      sampling (t, 1, NMarks, nsw, nFGL, prob); /* arg2 =1: EAT 3/2005 */

      if (!((t + 1) % ScorIt))
      {
/* now scoring will be as a fuzzy-p dsn for each locus and the max   */
/* If t(S^{(j)}) is exceeded by exactly k of the t(S_0^{(i)}) and tied with w,
 * then contrib to the histogram is rectangle area 1/n, and base endpoints
 *    k/(m+1), (k+1+w)/(m+1)
 */
         score_all (scoraff, fzpval, NMarks, nnull, qndsn, mean_stat);

/* scores log-segs and log-pens       */
         for (c = 0; c < n_compnts; ++c)
         {
            full_seg_prob (NMarks, c, recounts, temp_log_prob_seg);
            for (j = 1; j < NMarks; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
            }
            tot_log_prob_pen[c][NMarks] += log (prob[c][NMarks]);
         }
         ++tot_count;
      }
   }
   printf ("\n");
   prntStatusFMC ();

   printf ("\n Values of expectation statistics, max and by locus\n");
   for (j = 0; j <= NMarks; ++j)
      printf (" %7.3f", mean_stat[j] / tot_count);

   for (j = 0; j <= NMarks; ++j)
      norcum (NIt, nnull, fzpval[j]);
   printf ("\n\n Combined distribution of fuzzy p-values, by locus:\n");
   printf ("pval maxim");
   pr_pvals (NMarks, fzpval, nnull, nnull / 10);

   for (c = 0; c < n_compnts; ++c)
      if (!cmplist[c + 1].naff)
         printf ("    There are no affected individuals for component %d\n \n",
                 c + 1);
      else
      {
         pr_cmpnthead (c);
         pr_log_prob_seg (1, NMarks, -cmplist[c + 1].nmeioses,
                          tot_log_prob_seg[c], tot_count);

         pr_log_prob_pen (1, NMarks, tot_log_prob_pen[c], tot_count);

         pr_reco_counts (1, NMarks, recounts[c]);

      }
   fflush (stdout);

   /* Free up allocated memory */

   free_xxgg ();

   free_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                &tot_log_prob_pen);
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (NMarks + 1, &prob);
   free_logprob ();
   free_msmpl ();
   free_traits ();
   free_cmplists ();
   free_ibd_measures ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (Scan)
      free_perm ();

   exit (EXIT_SUCCESS);
}

/** @} */
