/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "quant.h"

/*****************************************************************
  Function to form a data matrix for the pedigree.
  This version uses the "genetic data columns" as fixed-effects indicators,
  and also reads in a quantitative data value "quant" -- becomes "x"
  The nghds() and sortdat() routines are then called

  Various checks on pedigree size etc. are carried out. The global variables
  "pedsize" and "namestring" are also evaluated here.

   This new version installed in the Quant library; 8/31/91
   Modified for greater flexibility; 9/29/91 -- same routines now
                used by univar, bivar, and also by mixd

   Modified to use command-line specified data-file;  4/22/92.
   Modified 12/92 (SL) to read additional variables
   Modified 2/93 (EAT) to make compatible with other data inputs, again.
   Modified 6/93 (EAT) to transfer pedigree input to Nghds.
   Modified 8/96 (MAJ) for improved output data format
   Modified 9/04 (EAT) to store trait data in nghd[i].q_dat
   Modifed  5/05 (MAJ) input_data() is replaced with set_qdat()
                       and setQuantEff().  nghds() is no longer
                       called from here.

   set_qdat() is now the main function used by Autozyg and Lodscore
   setQuantEff() is the main function used by PolyEM

*****************************************************************/

static void sortdat (double **, int);
static void sortqdat (double **, int);
static void sortfixed (int);
static int missfixed (int);
static int printfixed (int);

/*****************************************************************/
/* Sorting subroutines for data and effects                      */
/*****************************************************************/

static void sortdat (double **qq, int nt)
{
   int   i, j, k, qi;

   for (i = 0; i <= NFix; ++i)
      IFix[i] = 1;
   for (i = 0; i < pedsize; ++i)
   {
      qi = 0;
      for (k = 0; k < nt; ++k)
         if ((int) qq[k][i] == MISSING)
            qi = MISSING;

/* Modified EAT July 2002: first check whether covariate info is missing */
      if (missfixed (i))
         qi = MISSING;

      if (qi != MISSING)
      {
         ++Nq;
         idex[Nq] = i;
         qdex[i] = Nq;
         for (j = 0; j < nt; ++j)
            x[j][Nq] = qq[j][i];
         sortfixed (i);
      }
      else
         qdex[i] = 0;
   }
   return;
}

static int missfixed (int i)
/* routine added by EAT July 2002, to check for missing covariate info */
{
   int   mm, j, k;
   mm = 0;

   if (FESex)
      k = 3;
   else
      k = 4;

   for (j = 1; j <= NFix; ++j)
   {
      if (data[i][j + k] == 0)
         mm = 1;
   }
   return (mm);
}

static void sortfixed (int i)
{                                       /* routine to set fixed effects for
                                         * Nq'th data individual; ith by name */
   int   j, k;

   if (FESex)
      k = 3;
   else
      k = 4;

   fixdex[0][Nq] = 1;                   /* overall mean effect */
   for (j = 1; j <= NFix; ++j)
   {
      fixdex[j][Nq] = data[i][j + k];
      if (fixdex[j][Nq] < 0)
         give_up (ERQ_nfe, FENo[j], nghd[i].self);
      else if (fixdex[j][Nq] > LEVELS)
         give_up (ERQ_rfe, LEVELS, FENo[j], nghd[i].self);
   }
   return;
}

static int printfixed (int nt)
{
   int   i, j, k;
   int   nf[LEVELS + 1];

   printf ("% d individuals are observed for", Nq);
   if (nt == 1)
      printf (" 1 quantitative trait");
   else if (nt > 1)
      printf (" %d quantitative traits", nt);
   if (NFix == 1)
      printf (" and 1 fixed effect");
   else if (NFix > 1)
      printf (" and %d fixed effects", NFix);
   printf ("\n\n");
   fflush (stdout);

   if (!Nq)
      return Nq;

   for (i = 1; i <= NFix; ++i)
      IFix[i] = intmax (fixdex[i], 1, Nq);

   if (NFix > 0)
   {
      printf (" fixed    number       number observed                  ");
      printf ("present in trait\n");
      printf (" effect   of levels    in each level                    ");
      for (j = 0; j < nt; ++j)
         printf ("%d   ", TrNo[j]);
      printf ("\n ------   ---------    ---------------                  ");
      printf ("----------------");
      for (j = 1; j <= NFix; ++j)
      {
         for (i = 1; i <= IFix[j]; ++i)
            nf[i] = 0;
         for (i = 1; i <= Nq; ++i)
            ++nf[fixdex[j][i]];
         printf ("\n %4d     %4d        ", FENo[j], IFix[j]);
         for (i = 1; i <= IFix[j]; ++i)
         {
            if (i != 1 && i - (i / 6) * 6 == 1)
               printf ("\n                      ");
            printf ("%5d", nf[i]);
            if ((i <= 6 && i == IFix[j]) || i == 6)
            {
               for (k = 1; k <= 6 - i; ++k)
                  printf ("     ");
               printf (" ");
               for (k = 0; k < nt; ++k)
                  if (FfDex[k][j] == 0)
                     printf ("   N");
                  else
                     printf ("   Y");
            }
         }
      }
      printf ("\n\n");
      fflush (stdout);
   }
   return Nq;
}

/*****************************************************************
NEW Routine to make changes on reorganization of trait data:
new routine used by lm_markers, locus_mjrpeel, ....., but not yet
by PolyEM programs.
EAT Aug 2004

Also sortdat() replaced (temporaily?) by sortqdat
 Sorting subroutines for data and effects                      
*****************************************************************/

void set_qdat (double **quant)
{
   int   j;

   /* Now allocate effects etc. , since pedsize now known          */
   /* ("PedSize" is maximum pedigree size, "pedsize" is size read) */

   if (NTraits > 0)
   {
      x = malloc (NTraits * sizeof (double *));
      alloc_chk (WHERE, x);
      for (j = 0; j < NTraits; ++j)
      {
         x[j] = malloc ((pedsize + 1) * sizeof (double));
         alloc_chk (WHERE, x[j]);
      }

      /* This often seems to be getting unassigned - SCH */
      for (j = 0; j < NTraits; j++)
         x[j][0] = 0.0;

      qdex = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, qdex);
      idex = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, idex);
      for (j = 0; j <= NFix; ++j)
      {
         fixdex[j] = malloc ((pedsize + 1) * sizeof (int));
         alloc_chk (WHERE, fixdex[j]);
      }

      /* Now sort the fixed effects  */
      sortqdat (quant, NTraits);
      if (!printfixed (NTraits))
         /* if no observed individuals were found, something is amiss */
         give_up_where (WHERE, ERQ_noi);

      /*
       * Now done with quant ; observations are in x; and idex will permit
       * restoring original indexing if necessary
       */

      /* x, idex and fixdex are remalloced for observations  */
      for (j = 0; j < NTraits; ++j)
      {
         x[j] = realloc (x[j], (Nq + 1) * sizeof (double));
         alloc_chk (WHERE, x[j]);
      }
      idex = realloc (idex, (Nq + 1) * sizeof (int));
      alloc_chk (WHERE, idex);

      for (j = 0; j <= NFix; ++j)
      {
         fixdex[j] = realloc (fixdex[j], (Nq + 1) * sizeof (int));
         alloc_chk (WHERE, fixdex[j]);
      }
   }
   return;
}

static void sortqdat (double **qq, int nt)
{
   int   i, j, k, qi, np1;
   double *pdat;

   for (i = 0; i <= NFix; ++i)
      IFix[i] = 1;

/*  First counts how many we have not-missing: EAT Aug 2004  */

   np1 = 0;
   for (i = 0; i < pedsize; ++i)
   {
      qi = 0;
      for (k = 0; k < nt; ++k)
         if ((int) qq[k][i] == MISSING)
            qi = MISSING;

/* Modified EAT July 2002: first check whether covariate info is missing */
      if (missfixed (i))
         qi = MISSING;

      if (qi != MISSING)
         ++np1;
   }

   Nq = 0;
   for (i = 0; i < pedsize; ++i)
   {
      qi = 0;
      for (k = 0; k < nt; ++k)
         if ((int) qq[k][i] == MISSING)
            qi = MISSING;

/* Modified EAT July 2002: first check whether covariate info is missing */
      if (missfixed (i))
         qi = MISSING;

      if (qi != MISSING)
      {
         if (Nq == 0)
         {
            pdat = malloc (NTraits * np1 * sizeof (double));
            alloc_chk (WHERE, pdat);
         }
         nghd[i].q_dat = pdat + NTraits * Nq;
         for (j = 0; j < NTraits; ++j)
            nghd[i].q_dat[j] = qq[j][i];
         nghd[i].obs = 1;
         ++Nq;
      }
   }
   return;
}

void setQuantEff (double **quant)
{
   int   j;

   /* Now allocate effects etc. , since pedsize now known          */
   /* ("PedSize" is maximum pedigree size, "pedsize" is size read) */

   if (NTraits > 0)
   {
      x = malloc (NTraits * sizeof (double *));
      alloc_chk (WHERE, x);
      for (j = 0; j < NTraits; ++j)
      {
         x[j] = malloc ((pedsize + 1) * sizeof (double));
         alloc_chk (WHERE, x[j]);
      }

      /* This often seems to be getting unassigned - SCH */
      for (j = 0; j < NTraits; j++)
         x[j][0] = 0.0;

      qdex = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, qdex);
      idex = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, idex);
      for (j = 0; j <= NFix; ++j)
      {
         fixdex[j] = malloc ((pedsize + 1) * sizeof (int));
         alloc_chk (WHERE, fixdex[j]);
      }

      /* Now sort the fixed effects  */
      sortdat (quant, NTraits);
      if (!printfixed (NTraits))
         /* if no observed individuals were found, something is amiss */
         give_up_where (WHERE, ERQ_noi);

      /*
       * Now done with quant ; observations are in x; and idex will permit
       * restoring original indexing if necessary
       */

      /* x, idex and fixdex are remalloced for observations  */
      for (j = 0; j < NTraits; ++j)
      {
         x[j] = realloc (x[j], (Nq + 1) * sizeof (double));
         alloc_chk (WHERE, x[j]);
      }
      idex = realloc (idex, (Nq + 1) * sizeof (int));
      alloc_chk (WHERE, idex);

      for (j = 0; j <= NFix; ++j)
      {
         fixdex[j] = realloc (fixdex[j], (Nq + 1) * sizeof (int));
         alloc_chk (WHERE, fixdex[j]);
      }
   }
   return;
}
